//
// Copyright 2016-2021 by Garmin Ltd. or its subsidiaries.
// Subject to Garmin SDK License Agreement and Wearables
// Application Developer Agreement.
//

import Toybox.Application.Storage;
import Toybox.Lang;
import Toybox.WatchUi;

class TzCache
{

var max;            // current size of the cache
var tz_name as Array<String> or Null;        // cache of time zone name
var tz_index as Array<Number> or Null;       // cache on index into the tz info array

//! Constructor
public function initialize()
{

    max = Storage.getValue("TZmax");
    if (max != null) {
        tz_name = Storage.getValue("TZname");
        tz_index = Storage.getValue("TZindex");
    } else {
        reset();
    }
}

public function get(idx) as Array<String or Number>
{

    var name = tz_name[idx];
    var index = tz_index[idx];
    return [ name, index ];
}

public function index(name)
{

    for (var i = 0; i < max; ++i) {
        if (name.equals(tz_name[i])) {
            return i; }
    }
    return -1;
}

public function delete(idx)
{

    if (idx > 0) {
        for (var i = idx; i < (max - 1); ++i) {
            tz_name[i] = tz_name[i + 1];
            tz_index[i] = tz_index[i + 1];
        }
        max -= 1;
    }
}

public function push(name, idx)
{

    if (name.equals("Local")) {
        return; }

    for (var i = 0; i < max; ++i) {
        if (tz_name[i].equals(name)) {
            delete(i);
            break;
        }
    }

    if (max < C.CACHE_MAX) {
        ++max; }

    for (var i = max - 2; i >= 1; --i) {
        tz_name[i + 1] = tz_name[i];
        tz_index[i + 1] = tz_index[i];
    }
    tz_name[1] = name;
    tz_index[1] = idx;

    save();
}

public function reset()
{

    tz_name = new [C.CACHE_MAX];
    tz_index = new [C.CACHE_MAX];

    max = 1;
    tz_name[0] = "Local";
    tz_index[0] = 0;

    save();
}

private function save()
{

    Storage.setValue("TZmax", max);
    Storage.setValue("TZname", tz_name);
    Storage.setValue("TZindex", tz_index);
}

public function prt(hdr)
{

    C.logger(hdr + ":" + max);
    for (var i = 0; i < max; ++i) {
        C.logger("  " + i + ":" + tz_name[i] + ", " + tz_index[i]);
    }
}

}

